#ifndef TESTS
module EdenSkel (parMap, --parMapL,
		 farm, farmND, 
		 workpool, workpoolD, 
		 dm, dmI
	) where
#endif

import Eden hiding (parMap, farm)
import List

-- Tests:

main = 	do 
	 --print (parMap (+1) [1..10::Int]) 
	 --print (farm noPe 2 (+1) [1..100::Integer]) 
	 --print (workpoolD 3 noPe 4 (+1) [1..100::Integer]) 
	 --print (workpool 3 noPe 4 (+1) [1..100::Integer]) 
	 --print (dm (repeat (+1)) [1..10::Int]) 
	 print (dmI 10 (\xs i -> ((xs!!i)::Int)+1)) 
	 --print "End"


-- the worker processes
-------------------------
workerC :: (Transmissible a, Transmissible b) 
					=> (a -> b) -> Process [[a]] [[b]]
workerC f = process xs -> map (map f) xs
	
worker :: (Transmissible a, Transmissible b) => (a -> b) -> Process [a] [b]
worker f = process xs -> map f xs

workerS :: (Transmissible a, Transmissible b) => (a -> b) -> Process a b
workerS f = process x -> f x



-- the skeletons
-------------------
parMap :: (Transmissible a, Transmissible b) => 
		(a -> b) -> [a] -> [b]
parMap f xs = 
	eagerInstList (repeat (process x -> f x)) xs
	--let insts = map (cpTL proc) xs
		--     in tlList insts


farm :: (Transmissible a, Transmissible b) => 
		Int -> Int -> (a -> b) -> [a] -> [b]
farm np chunk f inputs 
	= let runs = unshuffleN np inputs
              outRuns = if chunk == 1
			then eagerInstList (repeat (worker f)) runs
			else map concat (eagerInstList (repeat (workerC f)) (map (splitAtN chunk) runs))
	  in outRuns `seq` shuffleN outRuns

farmND :: (Transmissible a, Transmissible b) => 
		Int -> Int -> (a -> b) -> [a] -> [b]
farmND np chunk f inputs 
	= let runs = unshuffleN np inputs
              outRuns = if chunk == 1
			then eagerInstList (repeat (worker f)) runs
			else mergeProc (eagerInstList (repeat (workerC f)) (map (splitAtN chunk) runs))
	  in outRuns `seq` shuffleN outRuns


workpoolD :: (Transmissible a, Transmissible b) => 
		Int -> Int -> Int -> (a -> b) -> [a] -> [b]
workpoolD prefetch np chunk f inputs = result
  where
    result = 
#ifdef TESTDISTRIB
	     (trace (show (map (length) outputsChildren))) 
#endif
	     (results)
    pids = [ 0..np-1]
    outputsChildren = map concat aux
    aux = eagerInstList (repeat (workerC (\ (x,y) -> (x,f y)))) (inputss)


    -- Non-deterministic merge of the results!!
    results = (map snd) . (mergeRunsBy cmpFst)  $ outputsChildren  

    cmpFst :: (Int,a) -> (Int,a) -> Bool
    cmpFst (i,x) (j,y) = i < j  

    newReqs = mergeProc (zipWith ($) [map (\x -> i) | i <- pids] 
				     outputsChildren)
		 
       -- The list of works to be done (inputs), is divided into a 
       -- list of list (inputss), where each internal list denotes the 
       -- works to be performed by a worker. The first two works of each
       -- worker (initialInputs) are given deterministically. The rest
       -- are given non-deterministically, as each process finish its
       -- work.

    numbered_inputs = (splitAtN chunk) (zip [0..] inputs)
    inputss =  distribute np numbered_inputs (initReqs ++ newReqs)
    initReqs = (concat (replicate prefetch [0..np-1]))

distribute np tasks requests
	     = map (f requests tasks) [0..np-1]
     	     where   -- f allows a parallel distribution
		     f _ [] _ = []
		     f [] _ _ = []
		     f (r:rs) (t:ts) n 
			| r == n = t : (f rs ts n)
		     	| otherwise    = (f rs ts n)

workpool :: (Transmissible a, Transmissible b) => 
		Int -> Int -> Int -> (a -> b) -> [a] -> [b]
workpool prefetch np chunk f inputs = result
  where
    result = 
#ifdef TESTDISTRIB
	     (trace (show (map (length) outputsChildren))) 
#endif
	     (results)
    pids = [ 0..np-1]
    outputsChildren = map concat aux
    aux = eagerInstList (repeat (workerC f)) (inputss)


    -- Non-deterministic merge of the results!!
    results = (mergeProc) outputsChildren  

    newReqs = mergeProc (zipWith ($) [map (\x -> i) | i <- pids] 
				     outputsChildren)
		 
       -- The list of works to be done (inputs), is divided into a 
       -- list of list (inputss), where each internal list denotes the 
       -- works to be performed by a worker. The first two works of each
       -- worker (initialInputs) are given deterministically. The rest
       -- are given non-deterministically, as each process finish its
       -- work.

    inputss =  distribute np ((splitAtN chunk) inputs) (initReqs ++ newReqs)
    initReqs = (concat (replicate prefetch [0..np-1]))


dm :: (Transmissible a, Transmissible b) => [a -> b] -> [a] -> [b]
dm fs inputs 
  = eagerInstList (map workerS fs) inputs

dmI :: Transmissible b => Int -> (a -> Int -> b) -> a -> [b]
dmI n f dat = eagerInstList [process _ -> f dat i | i<- [0..n-1]] 
				(repeat void)


t__ = dummySymbols

tl__ = createProcessTL pid (1::Int)

{-! global: NFData !-}
