module Main where

#if defined(RANDOM)
import Random
#else
import Inputs11
#endif
import ModArithm
-- import CRA

import qualified GHC.Conc (numCapabilities)
#if defined(STRATEGIES)
# if defined(SM_STRATEGIES)
import Control.Parallel
import Control.DeepSeq
import Control.Parallel.Strategies
# elif defined(EVAL_STRATEGIES)
import Control.Parallel
import Control.Parallel.SeqStrategies  hiding (using, Done(..))
import Control.Parallel.Strategies hiding (concat)
#  else
import Control.Parallel
import Control.Parallel.Strategies hiding (parZipWith)
# endif
#else
import Control.Parallel
#endif

import System

-- based on alleged Mathematica implementation
cra :: [Integer] -> [Integer] -> (Integer, Integer)
cra xs ms = (foldl (modSum m') 0 (zipWith bin_cra xs ms), m')
  where bin_cra x m = x * b * m0 
                      where b =  modInv m m0
                            m0 = m' `div` m 
        m' = product ms

#if defined(STRATEGIES)
# if defined(EVAL_STRATEGIES)

parZipWith :: Strategy c -> (a -> b -> c) -> [a] -> [b] -> [c]
parZipWith strat f xs ys = zipWith f xs ys `using` parList strat

craPar :: Int -> [Integer] -> [Integer] -> (Integer, Integer)
craPar z xs ms = cra xs' ms' 
                 where (xs', ms') = unzip $ parZipWith (ins rnf) cra (kluster z xs) (kluster z ms)

                       -- worker :: [Integer] -> [Integer] -> (Integer, Integer)
                       -- worker xs ms = (foldl nplus 0 {- map (nfib . (+ 23) . (`mod` 3)) -} $ xs, product ms)

{-
parMapFold1 :: Strategy b -> (a -> b) -> (b -> b -> b) -> [a] -> b
parMapFold1 strat f g xs = foldl1 g (map f xs  `using` parList strat)

parMapFold1Chunked :: Strategy b -> Int -> (a -> b) -> (b -> b -> b) -> [a] -> b
parMapFold1Chunked strat z f g xs = foldl1 g ((map worker (kluster z xs))  `using` parList strat)
                                    where worker xs = foldl1 g (map f xs)

craPar :: Int -> [Integer] -> [Integer] -> (Integer, Integer)
-- craPar z xs ms = foldl1 (\ (x,m) (y,n) -> (modSum m' x y, m*n)) (craPar' z xs ms `using` parList (parPair (ins rnf) (ins rnf))) 
craPar z xs ms = cra xs' ms' 
                 where m' = product ms
                       (xs', ms') = unzip $ (craPar' z xs ms `using` parList (parPair (ins rnf) (ins rnf)) )

craPar' :: Int -> [Integer] -> [Integer] -> [(Integer, Integer)]
craPar' z xs ms | length xs < z = [cra xs ms]
craPar' z xs ms | otherwise = y : ys 
                  	      where (xs1, xs2) = splitAt z xs
                  	            (ms1, ms2) = splitAt z ms
                                    y  = cra xs1 ms1
                                    ys = craPar' z xs2 ms2                        
-}

# elif defined(SM_STRATEGIES)

parZipWith :: Strategy c -> (a -> b -> c) -> [a] -> [b] -> [c]
parZipWith strat f xs ys = zipWith f xs ys `using` parList strat

craPar :: Int -> [Integer] -> [Integer] -> (Integer, Integer)
craPar z xs ms = cra xs' ms'
                 where (xs', ms') = unzip $ parZipWith rdeepseq cra (chunk z xs) (chunk z ms)

chunk :: Int -> [a] -> [[a]]
chunk _n [] = []
chunk n  xs = ys:chunk n zs where (ys,zs) = splitAt n xs

# else 

parZipWith :: Strategy c -> (a -> b -> c) -> [a] -> [b] -> [c]
parZipWith strat f xs ys = zipWith f xs ys `using` parList strat

craPar :: Int -> [Integer] -> [Integer] -> (Integer, Integer)
craPar z xs ms = cra xs' ms'
                 where (xs', ms') = unzip $ parZipWith rnf cra (chunk z xs) (chunk z ms)

chunk :: Int -> [a] -> [[a]]
chunk _n [] = []
chunk n  xs = ys:chunk n zs where (ys,zs) = splitAt n xs

# endif
#endif

#if 0
-- stupid divide-and-conquer version
craParDnc :: [Integer] -> [Integer] -> (Integer, Integer)
craParDnc xs ms | length xs < 100 = cra xs ms
                | otherwise       = cra [y',y''] [p',p''] `using` \ res -> (ins rnf) y' `par` ((ins rnf) y'' `pseq` (ins rnf) res)
                                    where (xs',xs'') = splitAt (length xs `div` 2) xs
                                          (ms',ms'') = splitAt (length ms `div` 2) ms
                                          -- p'  = product ms'
                                          -- p'' = product ms''
                                          (y', p')   = craParDnc xs' ms'
                                          (y'', p'') = craParDnc xs'' ms''

#endif

main = do
     xs <- getArgs
     let n = (read (xs!!0)) :: Int
     -- let m = (read xs!!1) :: Int
#if defined(RANDOM)
     g <- newStdGen
     let m = 65535 :: Integer
     let ms' = randomRs (2^16::Integer,2^24) g
     let ms  = take n $ fst $ unzip $ filter (\ (m,p) -> gcd m p == 1) (zip ms' (scanl1 (*) (1:ms')))
     putStrLn $ "Sanity check: are all moduli rel prime? "++(if and [ gcd m m' == 1 | m <- ms, m' <- ms, m/=m' ] then " OK " else " WRONG ")
     g' <- newStdGen
     let xs' = randomRs (2^16::Integer,2^24) g' -- randomRs (0::Integer,m) g
     let xs =  take n $ zipWith mod xs' ms
     -- let res0 = snd (seq_list_CRA (product ms) ms xs (repeat 1))
     -- let res0 = snd (seq_list_CRA0 ms xs)
     -- putStrLn $ "CRA: "++(show res0)++"; "++(if (all (\ (r,x,m) -> r `mod` m == x) (zip3 (repeat res0) xs ms)) then " OK " else " WRONG ")
     -- let res = fst $ cra xs ms
     -- putStrLn $ "xs': "++(show (take n xs'))
     putStrLn $ "xs: "++(show xs)
     putStrLn $ "ms: "++(show ms)
     -- putStrLn $ "CRA: "++(show res)++"; "++(if (all (\ (r,x,m) -> r `mod` m == x) (zip3 (repeat res) xs ms)) then " OK " else " WRONG ")
#else
     let xs = if n>(length xs_caf_3)
                then error $ "Can only supply up to "++(show (length xs_caf_3))++" elements"
                else take n xs_caf_3
     let ms = if n>(length ms_caf_3)
                then error $ "Can only supply up to "++(show (length ms_caf_3))++" elements"
                else take n ms_caf_3
#if defined(STRATEGIES)
     let z = (length xs `div` (GHC.Conc.numCapabilities * 2 {- NB: generate (* 2) threads for good performance -}))
     putStrLn $ "Computing cra over fixed lists of length " ++ (show n) ++ " using parZipWith and kluster size " ++ (show z)
     let res = fst $ craPar z xs ms
#else
     putStrLn $ "Computing sequential CRA over fixed lists of length " ++ (show n) ++ " "
     let res = fst $ cra xs ms  -- new version; see above -- OK
     -- let res = snd $ seq_list_CRA0 ms xs -- old version from CRA.hs -- OK
#endif
     putStrLn $ (res `pseq` "Done")
     -- putStrLn $ "CRA: "{- ++(show res)++"; "-}++(if (all (\ (r,x,m) -> r `mod` m == x) (zip3 (repeat res) xs ms)) then " OK " else " WRONG ")

#endif

-- From CRA.hs
seq_list_CRA0 (m:ms) (r:rs) = 
  foldl (\ (m0,r0) (m1,r1) -> (m0*m1, binCRA m0 m1 (modInv m1 m0) r0 r1) )
        (m,r) 
        (zip ms rs)


-- Same as above but applicable for infinite lists

seq_list_CRA :: Integer -> [Integer] -> [Integer] -> [Integer] -> (Integer,Integer)
seq_list_CRA mBound (m:ms) (r:rs) (d:ds)  = seq_list_CRA' mBound ms rs ds m r 

seq_list_CRA' :: Integer -> [Integer] -> [Integer] -> [Integer] -> Integer -> Integer ->
	    (Integer,Integer)
seq_list_CRA' mBound (m:ms) (r:rs) (d:ds) accM accR 
  | accM > mBound = (accM, accR)
  | d == 0     	  = seq_list_CRA' mBound ms rs ds accM accR
  | otherwise  	  = -- trace ("seq_list_CRA': (m,accM) = " ++ (show (m,accM))) $
                         seq_list_CRA' mBound ms rs ds (m*accM) 
                                       (my_cra accM m (modInv m accM) accR r)
                         where my_cra = binCRA {- or par_binCRA -}

binCRA :: Integer -> Integer -> Integer -> Integer -> Integer -> Integer 
binCRA m1 m2 inv a1 a2 = (if d == 0 then a1
                                    else a)
                         where ab = modHom m2 a1
                               d  = modDif m2 a2 a1
                               b  = modProd m2 d inv
                               b0 = if (b+b>m2) then b-m2
                                                else b
                               a  =  m1*b0+a1             {- not HWL version -}

#if 0
nfib :: Integer -> Integer
nfib 0 = 1
nfib 1 = 1
nfib n = nfib (n-1) + nfib (n-2) + 1

-- stupid (+)
nplus :: Integer -> Integer -> Integer
nplus m n = if m<0 
             then error "nplus cannot handle neg first argument"
             else nplus' m n 0
            where nplus' 0 n acc = acc
                  nplus' m n acc = nplus' (m-1) n (acc+m)
#endif